import json
from pathlib import Path

from typer import Argument, Option, Typer

from an_at_sync.model import BaseEvent
from an_at_sync.program import Program, ProgramSettings


class Pipedream:
    def __init__(self, steps: dict) -> None:
        self.steps = steps


main = Typer()

ConfigOption = Option(
    "config.py",
    help="Config file to load",
    exists=True,
    file_okay=True,
    dir_okay=False,
    resolve_path=True,
)


@main.command("init")
def init():
    pass


sync = Typer()
main.add_typer(sync, name="sync")


@sync.command("events")
def events(config: Path = ConfigOption, sync_rsvps: bool = Option(False, "--rsvps")):
    config_file = Program.load_config(config)

    program = Program(
        settings=ProgramSettings(),
        activist_class=config_file.Activist,
        event_class=config_file.Event,
        rsvp_class=config_file.RSVP,
    )

    for event_result in program.sync_events():
        program.write_result(event_result)

        if (
            sync_rsvps
            and event_result.status != "failed"
            and isinstance(event_result.instance, BaseEvent)
        ):
            for rsvp_result in program.sync_rsvps_from_event(event_result.instance):
                program.write_result(rsvp_result)


@main.command("webhook")
def webhook(
    webhook_path: Path = Argument(
        ...,
        help="JSON file containing webhook body",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
    config: Path = ConfigOption,
):
    config_file = Program.load_config(config)

    with open(webhook_path) as f:
        trigger = json.load(f)

    pd = Pipedream(steps={"trigger": trigger})
    config_file.handler(pd)



import phonenumbers


def get_or_blank(dict_name, key):
    return dict_name.get(key, "")


def standardize_phone(phone_number: str):
    try:
        phone = phonenumbers.parse("+%s" % phone_number)
        return phonenumbers.format_number(
            phone, phonenumbers.PhoneNumberFormat.NATIONAL
        )
    except Exception:
        return phone_number


def convert_adr(adr):
    try:
        address_ln = (", ").join(get_or_blank(adr, "address_lines"))
    except Exception:
        address_ln = ""
    return (
        address_ln,
        get_or_blank(adr, "locality"),
        get_or_blank(adr, "region"),
        get_or_blank(adr, "postal_code"),
    )



from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import Iterable, List, Optional, Type

from pyairtable import Table as Airtable
from pydantic import BaseModel as PydanticModel
from pydantic import BaseSettings
from rich.console import Console
from typing_extensions import Literal

from an_at_sync.actionnetwork import ActionNetworkApi
from an_at_sync.model import (ActivistRepository, BaseActivist, BaseEvent,
                              BaseModel, BaseRSVP, EventRepository,
                              RSVPRepository)


class SyncResult(PydanticModel):
    status: Literal["unchanged", "inserted", "updated", "skipped", "failed"]
    kind: Literal["activist", "event", "rsvp", "webhook"]
    instance: Optional[BaseModel]
    e: Optional[Exception]

    class Config:
        arbitrary_types_allowed = True

    def __str__(self):
        return f"Status for {self.instance}: {self.status}" + (
            f" Exception: {self.e}" if self.e else ""
        )


class ProgramSettings(BaseSettings):
    an_api_key: str
    at_base: str
    at_activists_table: str
    at_events_table: str
    at_rsvp_table: str
    at_api_key: str

    class Config:
        env_file = ".env"


class Program:
    @staticmethod
    def load_config(config_path: Path):
        spec = spec_from_file_location("config", config_path)
        if spec is None or spec.loader is None:
            raise Exception("spec or spec.loader for config was None")
        module = module_from_spec(spec)
        spec.loader.exec_module(module)

        return module

    def __init__(
        self,
        settings: ProgramSettings,
        activist_class: Type[BaseActivist],
        event_class: Type[BaseEvent],
        rsvp_class: Type[BaseRSVP],
    ):
        an = ActionNetworkApi(api_key=settings.an_api_key)

        self.events = EventRepository(
            an=an,
            at=Airtable(
                settings.at_api_key, settings.at_base, settings.at_events_table
            ),
            klass=event_class,
        )
        self.activists = ActivistRepository(
            an=an,
            at=Airtable(
                settings.at_api_key, settings.at_base, settings.at_activists_table
            ),
            klass=activist_class,
        )
        self.rsvps = RSVPRepository(
            an=an,
            at=Airtable(settings.at_api_key, settings.at_base, settings.at_rsvp_table),
            klass=rsvp_class,
        )
        self.console = Console()

    def sync_activists(self) -> Iterable[SyncResult]:
        for activist in self.activists.all_from_actionnetwork():
            yield self.sync_activist(activist)

    def sync_activist(self, activist: BaseActivist) -> SyncResult:
        try:
            record = self.activists.get_airtable_record(activist)
            insert = record is None
            if insert:
                self.activists.insert_airtable_record(activist)
            else:
                if not self.activists.should_update_airtable_record(activist):
                    return SyncResult(
                        status="unchanged",
                        kind="activist",
                        instance=activist,
                        e=None,
                    )
                self.activists.update_airtable_record(activist)
            return SyncResult(
                status="inserted" if insert else "updated",
                kind="activist",
                instance=activist,
                e=None,
            )
        except Exception as e:
            return SyncResult(status="failed", kind="activist", instance=activist, e=e)

    def sync_events(self) -> Iterable[SyncResult]:
        for event in self.events.all_from_actionnetwork():
            yield self.sync_event(event)

    def sync_event(self, event: BaseEvent) -> SyncResult:
        try:
            record = self.events.get_airtable_record(event)
            insert = record is None
            if insert:
                self.events.insert_airtable_record(event)
            else:
                if not self.events.should_update_airtable_record(event):
                    return SyncResult(
                        status="unchanged",
                        kind="event",
                        instance=event,
                        e=None,
                    )
                self.events.update_airtable_record(event)
            return SyncResult(
                status="inserted" if insert else "updated",
                kind="event",
                instance=event,
                e=None,
            )
        except Exception as e:
            return SyncResult(status="failed", kind="event", instance=event, e=e)

    def sync_rsvps_from_event(self, event: BaseEvent):
        yield from self.sync_rsvps(
            self.rsvps.from_actionnetwork_for_event(
                event,
                events=self.events,
                activists=self.activists,
            ),
        )

    def sync_rsvps(self, rsvps: Iterable[BaseRSVP]):
        for rsvp in rsvps:
            yield self.sync_rsvp(rsvp)

    def sync_rsvp(self, rsvp: BaseRSVP):
        try:
            rsvp_record = self.rsvps.get_airtable_record(rsvp)
            insert = rsvp_record is None
            if insert:
                self.rsvps.insert_airtable_record(
                    rsvp,
                    activist_record=self.activists.get_airtable_record(rsvp.activist),
                    event_record=self.events.get_airtable_record(rsvp.event),
                )
            else:
                if not self.rsvps.should_update_airtable_record(rsvp):
                    return SyncResult(
                        status="unchanged",
                        kind="rsvp",
                        instance=rsvp,
                        e=None,
                    )
                self.rsvps.update_airtable_record(rsvp)
            return SyncResult(
                status="inserted" if insert else "updated",
                kind="rsvp",
                instance=rsvp,
                e=None,
            )
        except Exception as e:
            return SyncResult(status="failed", kind="event", instance=rsvp, e=e)

    def handle_webhook(self, webhook_body: List[dict]):
        for webhook_event in webhook_body:
            attendance = webhook_event.get("osdi:attendance")
            if attendance is None:
                yield SyncResult(status="skipped", kind="webhook")
                continue
            activist = self.activists.from_actionnetwork_url(
                attendance["_links"]["osdi:person"]["href"],
                custom_fields=attendance["_links"]["osdi:person"].get("custom_fields"),
            )
            yield self.sync_activist(activist)

            event = self.events.from_actionnetwork_url(
                attendance["_links"]["osdi:event"]["href"],
                custom_fields=attendance["_links"]["osdi:person"].get("custom_fields"),
            )
            yield self.sync_event(event)

            rsvp = self.rsvps.klass.from_actionnetwork(
                attendance,
                activist=activist,
                event=event,
                activist_record=self.activists.get_airtable_record(activist),
                event_record=self.events.get_airtable_record(event),
            )
            yield self.sync_rsvp(rsvp)

    def write_result(self, result: SyncResult):
        display_name = (" " + result.instance.display_name()) if result.instance else ""
        # TODO(mAAdhaTTah) convert to match when 3.10 is min version
        if result.status == "unchanged":
            self.console.print(":information:", end=" ")
            self.console.print(
                f"Syncing {result.kind}{display_name} resulted in no changes"
            )
        elif result.status == "inserted":
            self.console.print(":heavy_plus_sign:", end=" ")
            self.console.print(
                f"Syncing {result.kind}{display_name} inserted new model"
            )
        elif result.status == "updated":
            self.console.print(":white_check_mark:", end=" ")
            self.console.print(f"Syncing {result.kind}{display_name} succeeded")
        elif result.status == "failed":
            self.console.print(":x:", end=" ")
            self.console.print(
                f"Syncing {result.kind}{display_name} failed with error:"
            )
            self.console.print(result.e)
        elif result.status == "skipped":
            self.console.print(":information:", end=" ")
            self.console.print(f"{result.kind}{display_name} was skipped")
        else:
            raise Exception(f"Unhandled status {result.status}")



from __future__ import annotations

from abc import abstractmethod
from typing import (Any, Dict, Generic, Iterable, List, Optional, Type,
                    TypeVar, Union)

from pyairtable import Table
from pyairtable.formulas import match
from pydantic import BaseModel as PydanticModel

from an_at_sync.actionnetwork import ActionNetworkApi

T = TypeVar("T", bound="BaseModel")


class BaseModel(PydanticModel):
    @classmethod
    @abstractmethod
    def from_actionnetwork(cls: Type[T], source: dict, **kwargs: Any) -> T:
        raise NotImplementedError()

    @abstractmethod
    def display_name(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def pk(self) -> Dict:
        raise NotImplementedError()

    @abstractmethod
    def to_airtable(self) -> dict:
        raise NotImplementedError()

    def __str__(self):
        params = ", ".join(
            [f"{key}={value}" for key, value in self.to_airtable().items()]
        )
        return f"<{self.__class__.__name__} {params}>"


class BaseActivist(BaseModel):
    pass


class BaseRSVP(BaseModel):
    activist: BaseActivist
    event: "BaseEvent"

    @abstractmethod
    def activist_column(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def event_column(self) -> str:
        raise NotImplementedError()


class BaseEvent(BaseModel):
    def postprocess(self) -> Iterable[BaseEvent]:
        yield self

    def load_rsvps(
        self,
        an_event: dict,
        events: "EventRepository",
        rsvps: "RSVPRepository",
        activists: "ActivistRepository",
    ) -> Iterable[BaseRSVP]:
        for an_attendance in rsvps.an.get_attendances_from_event(an_event):
            person_url = an_attendance["_links"]["osdi:person"]["href"]
            activist = activists.from_actionnetwork_url(person_url)
            yield rsvps.klass.from_actionnetwork(
                an_attendance,
                event=self,
                activist=activist,
                event_record=events.get_airtable_record(self),
                activist_record=activists.get_airtable_record(activist),
            )


BaseRSVP.update_forward_refs()

M = TypeVar("M", bound="BaseModel")
KOD = TypeVar("KOD")
VOD = TypeVar("VOD")


class ObjectDict(Generic[KOD, VOD]):
    """
    This is obviously missing a lot to be a full-fledged dict
    but this works for what we want to do for now.
    """

    def __init__(self) -> None:
        self.keys: List[KOD] = []
        self.values: List[VOD] = []

    def __setitem__(self, key, val):
        for i, existing_key in enumerate(self.keys):
            if existing_key == key:
                self.values[i] = val
        else:
            self.keys.append(key)
            self.values.append(val)

    def get(self, key) -> Union[VOD, None]:
        for i, existing_key in enumerate(self.keys):
            if existing_key == key:
                return self.values[i]

        return None

    def __repr__(self):
        return f"<ObjectDict keys={str(self.keys)} values={str(self.values)}>"


class BaseRepository(Generic[M]):
    def __init__(self, an: ActionNetworkApi, at: Table, klass: Type[M]) -> None:
        self.an = an
        self.at = at
        self.klass = klass
        self.model_to_at = ObjectDict[M, dict]()
        self.at_to_model = ObjectDict[dict, M]()
        self.model_to_an = ObjectDict[M, dict]()
        self.an_to_model = ObjectDict[dict, M]()

    def get_airtable_record(self, model: M):
        at_model = self.model_to_at.get(model)
        if not at_model:
            at_model = self.at.first(formula=match(model.pk()))
            if at_model:
                self._associate_model(model, at_model=at_model)
        return at_model

    def insert_airtable_record(self, model: M, **kwargs):
        return self.at.create(model.to_airtable())

    def update_airtable_record(self, model: M):
        airtable_record = self.get_airtable_record(model)
        if airtable_record is None:
            raise Exception("Cannot update record that does not exist")
        return self.at.update(airtable_record["id"], model.to_airtable())

    def should_update_airtable_record(self, model: M):
        airtable_record = self.get_airtable_record(model)
        if airtable_record is None:
            raise Exception("Cannot update record that does not exist")
        fields = airtable_record["fields"]
        update = model.to_airtable()

        def empty_str_to_none(val):
            return None if val == "" else val

        # Handle situations where an empty string on the AN side maps
        # to None on the AT side, such that an update would produce
        # no changes but this would other mark them as different
        return {key: empty_str_to_none(update.get(key)) for key in update} != {
            key: empty_str_to_none(fields.get(key)) for key in update
        }

    def from_actionnetwork_url(self, url: str, **kwargs) -> M:
        an_model = self.an.session.get(url).json()
        model = self.an_to_model.get(an_model)
        if not model:
            model = self.klass.from_actionnetwork(an_model, **kwargs)
            self._associate_model(model, an_model=an_model)
        return model

    def _associate_model(
        self,
        model: M,
        *,
        an_model: Optional[dict] = None,
        at_model: Optional[dict] = None,
    ):
        if an_model:
            self.an_to_model[an_model] = model
            self.model_to_an[model] = an_model

        if at_model:
            self.at_to_model[at_model] = model
            self.model_to_at[model] = at_model


class EventRepository(BaseRepository[BaseEvent]):
    def all_from_actionnetwork(self) -> Iterable[BaseEvent]:
        for an_event in self.an.get_all_events():
            event = self.an_to_model.get(an_event)
            if not event:
                event = self.klass.from_actionnetwork(an_event)
                self._associate_model(event, an_model=an_event)
            yield from event.postprocess()


class ActivistRepository(BaseRepository[BaseActivist]):
    def all_from_actionnetwork(self) -> Iterable[BaseActivist]:
        for an_activist in self.an.get_all_activists():
            activist = self.an_to_model.get(an_activist)
            if not activist:
                activist = self.klass.from_actionnetwork(an_activist)
                self._associate_model(activist, an_model=an_activist)
            yield activist


class RSVPRepository(BaseRepository[BaseRSVP]):
    def insert_airtable_record(self, model: BaseRSVP, **kwargs):
        return self.at.create(
            {
                **model.to_airtable(),
                model.activist_column(): [kwargs["activist_record"]["id"]],
                model.event_column(): [kwargs["event_record"]["id"]],
            }
        )

    def from_actionnetwork_for_event(
        self, event: BaseEvent, events: EventRepository, activists: ActivistRepository
    ) -> Iterable[BaseRSVP]:
        an_event = events.model_to_an.get(event)
        if an_event is None:
            raise Exception(
                "Event passed to RSVPRepository#from_actionnetwork_for_event was not loaded from AN;"
                "Loading from AN for AT source not yet supported"
            )

        yield from event.load_rsvps(
            an_event, events=events, rsvps=self, activists=activists
        )



from typing import Iterable
from urllib.parse import quote

from requests import Session


class ActionNetworkApi:
    """
    Python wrapper for Action Network API.

    Cribbed primarily from https://github.com/PhillyDSA/pyactionnetwork
    with some added methods & typing.
    """

    def __init__(self, api_key, **kwargs):
        """Instantiate the API client and get config."""
        self.session = Session()
        self.session.headers.update({"OSDI-API-Token": api_key})
        self.refresh_config()
        self.base_url = self.config.get("links", {}).get(
            "self", "https://actionnetwork.org/api/v2/"
        )

    def refresh_config(self):
        """Get a new version of the base_url config."""
        self.config = self.session.get(url="https://actionnetwork.org/api/v2/").json()

    def resource_to_url(self, resource):
        """Convert a named endpoint into a URL.
        Args:
            resource (str):
                resource name (e.g. 'links', 'people', etc.)
        Returns:
            (str) Full resource endpoint URL.
        """
        if resource in self.config.get("_links", {}).keys():
            return self.config["_links"][resource]["href"]
        try:
            return self.config["_links"]["osdi:{0}".format(resource)]["href"]
        except KeyError:
            raise KeyError("Unknown Resource %s", resource)

    def get_resource(self, resource):
        """Get a resource endpoint by name.
        Args:
            resource (str):
                Resource endpoint of the format 'people', 'events', 'lists', etc.
        Returns:
            (dict) API response from endpoint or `None` if not found/valid.
        """
        url = self.resource_to_url(resource)
        return self.session.get(url).json()

    def get_person(self, person_id=None, search_by="email", search_string=None):
        """Search for a user.
        Args:
            search_by (str):
                Field by which to search for a user. 'email' is the default.
            search_string (str):
                String to search for within the field given by `search_by`
        Returns:
            (dict) person json if found, otherwise `None`
        """
        if person_id:
            url = "{0}people/{1}".format(self.base_url, person_id)
        else:
            url = "{0}people/?filter={1} eq '{2}'".format(
                self.base_url, search_by, quote(search_string)
            )

        resp = self.session.get(url)
        return resp.json()

    def create_person(
        self,
        email=None,
        given_name="",
        family_name="",
        address=list(),
        city="",
        state="",
        country="",
        postal_code="",
        tags=list(),
        custom_fields=dict(),
    ):
        """Create a user.
        Documentation here: https://actionnetwork.org/docs/v2/person_signup_helper
        Args:
            email ((str, list)):
                email address (or, if list, addresses) of the person
            given_name (str, optional):
                first name of the person
            family_name (str, optional):
                last name of the person
            address ((str, list), optional):
                address of the person. if a str, then one address line
                only. if a list, then address_lines in action network
                will be respected (for apartments or companies etc.)
            city (str, optional):
                city of the person.
            country (str, optional):
                country code for the person.
            postal_code (str, optional):
                postal or zip code of the person.
            tags ((str, list), optional):
                add any tags you want when creating a person.
            custom_fields (dict, optional):
                dict of custom fields to pass to the api
        Returns:
            (dict) A fully fleshed out dictionary representing a person,
            containing the above attributes and additional attributes
            set by Action Network.
        """
        url = "{0}people/".format(self.base_url)
        payload = {
            "person": {
                "family_name": family_name,
                "given_name": given_name,
                "postal_addresses": [
                    {
                        "address_lines": list(address),
                        "locality": city,
                        "region": state,
                        "country": country,
                        "postal_code": postal_code,
                    }
                ],
                "email_addresses": [{"address": email}],
                "custom_fields": custom_fields,
            },
            "add_tags": list(tags),
        }

        resp = self.session.post(url, json=payload)
        return resp.json()

    def update_person(
        self,
        person_id=None,
        email=None,
        given_name=None,
        family_name=None,
        address=list(),
        city=None,
        state=None,
        country=None,
        postal_code=None,
        tags=list(),
        custom_fields=dict(),
    ):
        """Update a user.
        Args:
            email ((str, list)):
                email address (or, if list, addresses) of the person
            given_name (str, optional):
                first name of the person
            family_name (str, optional):
                last name of the person
            address ((str, list), optional):
                address of the person. if a str, then one address line
                only. if a list, then address_lines in action network
                will be respected (for apartments or companies etc.)
            city (str, optional):
                city of the person.
            country (str, optional):
                country code for the person.
            postal_code (str, optional):
                postal or zip code of the person.
            tags ((str, list), optional):
                add any tags you want when creating a person.
            custom_fields (dict, optional):
                dict of custom fields to pass to the api
        Returns:
            (dict) A fully fleshed out dictionary representing a person, containing the above
            attributes and additional attributes set by Action Network.
        """
        url = "{0}people/{1}".format(self.base_url, person_id)
        payload = {
            "family_name": family_name,
            "given_name": given_name,
            "postal_addresses": [
                {
                    "address_lines": list(address),
                    "locality": city,
                    "region": state,
                    "country": country,
                    "postal_code": postal_code,
                }
            ],
            "email_addresses": [{"address": email}],
            "add_tags": list(tags),
            "custom_fields": custom_fields,
        }

        return self.session.put(url, json=payload).json()

    def get_all_activists(self) -> Iterable[dict]:
        url = self.resource_to_url("people")
        while url:
            body = self.session.get(url).json()
            for activist in body.get("_embedded", {}).get("osdi:people", []):
                yield activist

            next_link = body["_links"].get("next")
            url = next_link.get("href") if next_link else None

    def get_all_events(self):
        url = self.resource_to_url("events")
        while url:
            body = self.session.get(url).json()
            for event in body.get("_embedded").get("osdi:events"):
                yield event

            next_link = body["_links"].get("next")
            url = next_link.get("href") if next_link else None

    def get_event(self, event):
        url = self.resource_to_url("events")
        return self.session.get(f"{url}/{event}").json()

    def get_attendances_from_event(self, event: dict):
        next_href = event["_links"]["osdi:attendances"]["href"]

        while next_href:
            attendances_body = self.session.get(next_href).json()
            yield from attendances_body["_embedded"]["osdi:attendances"]

            try:
                next_href = attendances_body["_links"]["next"]["href"]
            except KeyError:
                next_href = None



#!/usr/bin/env python

# Always prefer setuptools over distutils
# To use a consistent encoding
from codecs import open
from os import path

from setuptools import setup

# The directory containing this file
HERE = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(HERE, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(HERE, "requirements.txt"), encoding="utf-8") as f:
    install_requires = f.read().split("\n")


setup(
    name="an_at_sync",
    version="0.4.0",
    description="Python package & cli for syncing between ActionNetwork & AirTable",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="James DiGioia",
    author_email="jamesorodig@gmail.com",
    url="https://github.com/mAAdhaTTah/an-at-sync",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Operating System :: OS Independent",
    ],
    packages=["an_at_sync"],
    install_requires=install_requires,
    entry_points={
        "console_scripts": ["an_at_sync=an_at_sync.cli:main"],
    },
)
